/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* 
 * Copyright (C) 2019 by Giam Teck Choon
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

/*!
 *  @file make_seekable.c
 *  @date 21 Sep 2019
 *  @author Giam Teck Choon
 *  @brief Function for make_seekable()
 *  @bug No known bugs.
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <bglibs/str.h>
#include <bglibs/iobuf.h>
#include "is_seekable.h"
#include "str_random.h"
#include "is_dir.h"
#include "is_file.h"

/*!
 *  @brief If file descriptor fd is seekable, we just return the file
 *         descriptor fd. Otherwise we create a temp. one and copy fd to
 *         newfd, dup2 newfd to fd and return fd instead.
 *  @param fd File descriptor.
 *  @return fd on success, or -1 on error and errno is set.
 */
int make_seekable(int fd)
{
	if (is_seekable(fd)) {
		return fd;
	}
	str rs = {0,0,0};
	if (!str_random(&rs, 16)) {
		if (!errno) errno = ENOMEM;
		str_free(&rs);
		return -1;
	}
	const char *tmp;
	tmp = getenv("TMP"); /* <stdlib.h> */
	if (!tmp || *tmp == 0) { /* $TMP not set or empty */
		if (is_dir("./tmp")) {
			tmp = "./tmp";
		} else {
			tmp = "/tmp";
		}
	}
	if (!is_dir(tmp)) {
		str_free(&rs);
		errno = ENOENT;
		return -1;
	}
	str t = {0,0,0};
	if (!str_copy3s(&t, tmp, "/", rs.s)) {
		str_free(&rs);
		str_free(&t);
		errno = ENOMEM;
		return -1;
	}
	str_free(&rs);
	if (is_file(t.s)) {
		str_free(&t);
		errno = EEXIST;
		return -1;
	}
	/*
	 * open needs <sys/types.h>, <sys/stat.h> and <fcntl.h>
	 */
	int newfd = open(t.s, O_RDWR|O_CREAT|O_EXCL, 0600);
	if (newfd < 0) {
		str_free(&t);
		return -1;
	}
	/*
	 * The following makes the temporary file disappear immediately
	 * on program exit by removing the only filename that links to
	 * it.
	 */
	if (unlink(t.s) != 0) { /* <unistd.h> */
		str_free(&t);
		return -1;
	}
	str_free(&t);
	/*
	 * Copy content of fd into newfd
	 */
	ibuf ib;
	ibuf_init(&ib, fd, 0, 0, 0);
	ibuf_copytofd(&ib, newfd);
	ibuf_close(&ib);
	/*
	 * lseek needs <sys/types.h> and <unistd.h>
	 */
	if (lseek(newfd, 0L, SEEK_SET) != 0) {
		return -1;
	}
	if (close(fd) || dup2(newfd, fd) || close(newfd)) {
		return -1;
	}
	return fd;
}
