/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* 
 * Copyright (C) 2019 by Giam Teck Choon
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

/*!
 *  @file qmail_virtualdomains.c
 *  @date 21 Sep 2019
 *  @author Giam Teck Choon
 *  @brief Function for qmail_virtualdomains()
 *  @bug No known bugs.
 */

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include <bglibs/dict.h>
#include "common.h"
#include "is_file.h"

static dict vdomains;
static str qmailcontroldir = {0,0,0};
static str vdpath = {0,0,0};

/*!
 *  @brief lowercase struct str. This is static function.
 *  @param s Convert struct str s to lowercase.
 *  @return On success, 1 is returned.
 */
static int map_lower(str *s)
{
	str_lower(s);
	return 1;
}

/*!
 *  @brief Free all static variables. This is static function.
 *  @return Void.
 */
static void _freemem(void)
{
	if (qmailcontroldir.s) str_free(&qmailcontroldir);
	if (vdpath.s) str_free(&vdpath);
	dict_free(&vdomains, dict_str_free);
}

/*!
 *  @brief Lookup system user from domain.
 *  @param domain Domain name in struct str.
 *  @param user System user of domain in struct str.
 *  @return On success, 1 is returned otherwise 0 is returned. On error, -1 is
 *          returned, and errno is set appropriately.
 */
int qmail_virtualdomains(str *domain, str *user)
{
	const char *tmp = getenv("QMAIL_CONTROL_DIR");
	if (tmp && tmp[0]) {
		if (!str_copys(&qmailcontroldir, tmp)) {
			errno = ENOMEM;
			_freemem();
			return -1;
		}
	} else {
		if (!str_copys(&qmailcontroldir, "/var/qmail/control")) {
			errno = ENOMEM;
			_freemem();
			return -1;
		}
	}
	if (!str_copy2s(&vdpath, qmailcontroldir.s, "/virtualdomains")) {
		errno = ENOMEM;
		_freemem();
		return -1;
	}
	if (!is_file(vdpath.s)) {
		errno = ENOENT;
		_freemem();
		return -1;
	}
	if (!dict_load_map(&vdomains, vdpath.s, 0, ':', map_lower, 0)) {
		errno = ENOMEM;
		_freemem();
		return -1;
	}
	str_lower(domain);
	dict_entry *e;
	if ((e = dict_get(&vdomains, domain)) == 0) {
		unsigned i;
		while ((i = str_findnext(domain, '.', 1)) != (unsigned)-1) {
			str_lcut(domain, i);
			if ((e = dict_get(&vdomains, domain)) != 0) {
				break;
			}
		}
	}
	if (e) {
		if (!str_copy(user, (str*)e->data)) {
			errno = ENOMEM;
			_freemem();
			return -1;
		}
	}
	_freemem();
	if (user->s && user->len > 0) {
		return 1;
	}
	return 0;
}
